/*
** Copyright (C) 1998 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "xalloca.h"
#include <stdio.h>
#if STDC_HEADERS
#  include <stdlib.h>
#  include <string.h>
#else
# if HAVE_STRING_H
#   include <string.h>
# else
#   include <strings.h>
# endif
#endif

#include "dstring.h"
#include "fmtstring.h"
#include "xmalloc.h"

/*
 * Create a new formatted dynamic string
 * left  -> Every time we enter a new line insert 'left' spaces
 * right -> Do not allow a word to go beyond column 'right'
 * wrap  -> If = 0, then truncate words that go over the margin
 *          If > 0, then wrap words that go over the margin to 
 *          column 'wrap'
 */
struct fmtstring *new_fmtstring (int left, int right, int wrap)
{
  /* Allocate the formatted dynamic string */
  struct fmtstring *fs = MALLOC (struct fmtstring, 1);

  /* Set the margins to their initial values */
  fs->left = left;
  fs->right = right;
  fs->wrap = wrap;

  /* Initialize line and column trackers */
  fs->line = 0;
  fs->column = 0;

  /* Initialize the wrap flag to 0 */
  fs->wrap_flag = 0;

  /* Set up the dynamic string */
  fs->string = MALLOC (dynamic_string, 1);
  ds_init (fs->string,5);

  /* Return the newly allocated structure fs */
  return fs;
}

/* Delete a formatted string */
void 
delete_fmtstring (struct fmtstring *fs)
{
  ds_destroy (fs->string);
  free (fs->string);
  free (fs);
}

/* Change the left margin */
int 
fmts_set_left (struct fmtstring *fs, int left)
{
  int oldleft = fs->left;
  fs->left = left;
  return oldleft;
}

/* Change the right margin */
int 
fmts_set_right (struct fmtstring *fs, int right)
{
  int oldright = fs->right;
  fs->right = right;
  return oldright;
}

/* Change the wrap */
int 
fmts_set_wrap (struct fmtstring *fs, int wrap)
{
  int oldwrap = fs->wrap;
  fs->wrap = wrap;
  return oldwrap;
}

/* Get the current line */
int 
fmts_get_line (struct fmtstring *fs)
{
  return fs->line;
}

/* Get the current column */
int 
fmts_get_column (struct fmtstring *fs)
{
  return fs->column;
}

/* Insert a newline */
static void 
fmts_newline (struct fmtstring *fs)
{
  ds_strcat (fs->string,"\n");
  fs->column = 0;
  (fs->line)++;
}

/* Do a line break. This takes us to a new line, move up to the left
   margin, and reset the wrap flag. */
void 
fmts_linebreak (struct fmtstring *fs)
{
  fmts_newline (fs);
  fmts_move_to_column (fs, fs->left);
  fs->wrap_flag = 0;
}


/* Move to column 'column'. If not possible, then go to that
   column on the next line */
void 
fmts_move_to_column (struct fmtstring *fs, int column)
{
  /* If we are already there, then do nothing */
  if (fs->column == column) return;

  /* If we can get there on this line, then do so */
  if (fs->column < column)
    {
      int i = column - (fs->column);
      while (i--) ds_strcat (fs->string, " ");
      fs->column = column;
    }
  /* otherwise, get a new line and try again */
  else
    {
      fmts_newline (fs);
      fmts_move_to_column (fs, column);
    }
  return;
}

/* This routine dumps the text between two pointer locations to the
   dynamic string */
static void 
fmts_dump_text (struct fmtstring *fs, char *p1, char *p2)
{
  /* p2-p1+1 is the length of the substring that we want to dump.
     Add one more byte for the terminating null */
  size_t length = ((size_t) ((p2-p1)+1));

  /* Copy the substring to temporary stack space. Don't forget termination */
  char *tmp = ALLOCA (char, (length+1));
  memcpy (tmp,p1,length);
  tmp[length] = '\0';

  /* If we are in behind the left margin and the wrap flag has not been set
     then move on to the left margin */
  if (fs->wrap_flag == 0 && fs->column < fs->left)
    fmts_move_to_column (fs, fs->left);

  /* Add to the dynamic string and update the column */
  ds_strcat (fs->string, tmp);
  fs->column += length;
}

/*
 * Catenate text to a string
 * *************************
 *
 * This is the heart of formatted data strings. This routine takes unformatted
 * text, formats it and inserts it to the dynamic string. We implement this
 * routine as a state machine. 
 *
 * The state machine maintains two variables which are character pointers
 * pointing inside the text that we want to format:
 * p1 = The beginning of a substring that we want to insert
 * p2 = The end of a substring that we want to insert
 * 
 * The state machine can be in one of the following states:
 * 1. Skipping whitespace
 * 2. Scanning a word
 * 3. Inserting the word between (p1,p2)
 * 4. Wrapping over to the next line
 * We now describe these three states in more detail
 *
 * --> Initialization: Set p1 = txt and p2 = txt. Goto state 1
 * 
 * 1. Skipping whitespace
 *    -------------------
 * 
 * Upon entry, we require that p1 == p2
 * If the current character is ' ' or '\n' 
 *     Increase both p1 and p2.
 *     goto (1)
 * else if the current character is '\v'
 *     Insert a newline. 
 *     Go to left margin.
 *     goto (1)
 * else if the current character is '\0'
 *     quit
 * else 
 *     goto (2)
 * end if
 *
 * 2. Scanning a word
 *    ---------------
 *
 * Upon entry, we require that p1 == p2
 * If p2 is ' ' or '\n' or '\v' or '\0'
 *     goto (3)
 * else
 *     Increase p2
 *     goto (2)
 * endif
 *
 * 3. Inserting the word
 *    ------------------
 *
 * Upon entry, (p1,p2) enclose the substring that we want to insert.
 * This substring is one word. p2 points at the end of that word including
 * one whitespace character
 * If the size of the word does not fit on this line
 *    Get a newline
 *    Go to the wrap margin
 * endif
 * Insert the text between p1, p2-1
 * Insert one whitespace
 * Update the column register
 * Set p1 = p2
 * goto (1)  (all necessary checking will be done in (1))
 * 
 */

void
fmts_catenate (struct fmtstring *fs, char *txt)
{
  /* Initialization */
  char *p1 = txt;
  char *p2 = txt;

 skip_whitespace:
  {
    if (*p1 == ' ' || *p1 == '\n')
      {
	p1++;
	p2++;
	goto skip_whitespace;
      }
    else if (*p1 == '\v')
      {
	fmts_linebreak (fs);
	goto skip_whitespace;
      }
    else if (*p1 == '\0')
      {
	return;
      }
    else
      {
	goto scan_word;
      }
  }
  
 scan_word:
  {
    if (*p2 == ' ' || *p2 == '\n' || *p2 == '\v' || *p2 == '\0')
      {
	goto insert_word;
      }
    else
      {
	p2++;
	goto scan_word;
      }
  }

 insert_word:
  {
    size_t length = p2 - p1 + 1;

    if (fs->column + ((int) length) >= fs->right)
      {
	fmts_newline (fs);
	fmts_move_to_column (fs, fs->wrap);
	fs->wrap_flag = 1;
      }
    fmts_dump_text (fs, p1, p2-1);
    ds_strcat (fs->string, " ");
    fs->column++;
    p1 = p2;
    goto skip_whitespace;
  }
}

/* 
 * Catenate text with more arguments 
 */

void fmts_catenate2 (struct fmtstring *fs, char *t1, char *t2)
{
  char *txt = ALLOCA (char, (strlen (t1) + strlen (t2) + 1));
  strcpy (txt,t1); strcat (txt,t2);
  fmts_catenate (fs,txt);
}

void fmts_catenate3 (struct fmtstring *fs, char *t1, char *t2, char *t3)
{
  char *txt = ALLOCA (char, (strlen (t1) + strlen (t2) + strlen (t3) + 1));
  strcpy (txt,t1); strcat (txt,t2); strcat (txt,t3);
  fmts_catenate (fs,txt);
}

void fmts_catenate4 (struct fmtstring *fs, char *t1, char *t2, char *t3,
		     char *t4)
{
  char *txt = ALLOCA (char, (strlen (t1) + strlen (t2) + strlen (t3) +
			     strlen (t4) + 1));
  strcpy (txt,t1); strcat (txt,t2); strcat (txt,t3); strcat (txt,t4);
  fmts_catenate (fs,txt);
}

/*
 * Catenate text that is provided as a substring that begins at 'begin'
 * and ends at 'end'
 */

void fmts_catenate_substring (struct fmtstring *fs, char *begin, char *end)
{
  size_t length = ((size_t) (end - begin + 1));
  char *txt = ALLOCA (char , length + 1);
  memcpy (txt, begin, length);
  txt[length] = '\0';
  fmts_catenate (fs, txt);
}

/*
 * Return the built string
 */

char *
fmts_string (struct fmtstring *fs)
{
  return ds_get_string(fs->string);
}

