/*
** Copyright (C) 1998, 1999 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <xalloca.h>
#include <stdio.h>
#include "xstdlib.h"
#include "xstring.h"
#include "xmalloc.h"
#include "dstring.h"
#include <getopt.h>
#include "parg.h"
#include "fmtstring.h"
#include <assert.h>

char *xstrdup (char *string);

/* The following are hooks for the standard --help, --version and
 * --usage options */
void parg_hook_usage (struct parg *parser, char *arg);
void parg_hook_help (struct parg *parser, char *arg);
void parg_hook_version (struct parg *parser, char *arg);

/*
** ==================
** Global variables 
** ==================
*/

/*
 * If set by the program, then program will tell user where person
 * should send per bug reports
 */

char *parg_program_bug_address = 0;

/*
 * These should be assigned equal to PACKAGE and VERSION as provided
 * by Autoconf. These assignments are mandatory if you like to have
 * the --version and --help options.
 */

char *parg_package_name = 0;
char *parg_package_version = 0;

/*
 *  If the individual program has a different name than the package
 * then assign the name of the program to 'parg_program_name'.
 * If the program has its own version number then assign it to
 * 'parg_program_version'. Both of these assignments are optional
 */

char *parg_program_name = 0;
char *parg_program_version = 0;

/*
 * If 'parg_version_hook' is nonzero, then it is invoked during --version
 * It can be used to print a short notice indicating that the software
 * is free software
 */

void (*parg_version_hook) (void) = 0;

/*
** Program documentation that must be printed before and after the options.
** The two parts are separated with \v. Either part is optional
*/

char *parg_program_documentation = 0;

/* 
** If nonzero, a string that describes the arguments accepted by the
** program. If separated with newlines, then alternative usage patterns
** are explained, according to the GNITS standards
*/

char *parg_program_arguments = 0;

/*
** If the system does not support the GNU program invocation variables
** then introduce them here
*/

#ifndef HAVE_PROGRAM_INVOCATION_SHORT_NAME
char *program_invocation_short_name;
#endif

#ifndef HAVE_PROGRAM_INVOCATION_NAME
char *program_invocation_name;
#endif

/*
** =======================================================================
** =======================================================================
*/

/*
 * -------------------------------------------------------------------
 * This routine must be called from main as soon as possible so that
 * the program can identify it's invocation information. 
 * A typical call should look like:
 *   parg_set_invocation (argv[0], PACKAGE, VERSION);
 * -------------------------------------------------------------------
 */

void parg_set_invocation (char *argv0, char *package, char *version)
{
  /* Look for the last '/' character */
  char *argv0_end = argv0 + ((size_t) (strlen (argv0) - 1));
  while (argv0_end != argv0 && *argv0_end != '/') argv0_end--;

  /* Now set the invocation variables */
  program_invocation_name = argv0;
  if (argv0_end == argv0)
    {
      program_invocation_short_name = program_invocation_name;
    }
  else
    {
      program_invocation_short_name = argv0_end + 1;
    }

  /* Set the package name and version */
  parg_package_name = xstrdup (package);
  parg_package_version = xstrdup (version);
}


/*
** ======================================================================
** ======================================================================
*/

/*
** The following functions handle the command-line switch parsing
*/

/* Return true if `opt' is the terminating option. Otherwise return false */
static int 
parg_terminating_option_p (struct parg_option *opt)
{
  return (((opt->name == 0) && (opt->key == 0) && (opt->arg == 0) &&
	   (opt->doc == 0) && (opt->hook == 0)) ? 1 : 0);
}

/* Return true if `opt' is a documentation option. Otherwise return true */
static int 
parg_documentation_option_p (struct parg_option *opt)
{
  return (((opt->name != 0) && (opt->key == 0)) ? 1 : 0);
}

/* Return true if `opt' defines a long option */
static int 
parg_long_option_p (struct parg_option *opt)
{
  return (((opt->key != 0) && (opt->name != 0)) ? 1 : 0);
}

/* Return true if `opt' defines a short option */
static int
parg_short_option_p (struct parg_option *opt)
{
  return (((opt->key != ':') && (opt->key != 0)) ? 1 : 0);
}

/* Return true if `opt' is an alias for another option.
   Beware! If opt->hook == 0, the option may still be a documentation 
   option! We do NOT want that! */
static int
parg_option_is_alias_p (struct parg_option *opt)
{
  return (((opt->hook == 0) && 
	   (parg_documentation_option_p (opt) == 0)) ? 1 : 0);
}

/* If an option is an alias, resolve it by returning an option that is
   not an alias */
static struct parg_option *
parg_resolve_alias (struct parg *parser, struct parg_option *opt)
{
 tail_recurse:
  if (parg_option_is_alias_p (opt))
    {
      /* Check whether we have an unresolvable alias */
      /* FIXME: I need to see how to throw errors like this with error */
      if (opt == parser->options)
	{
	  error (1, 0, "Fatal: Unresorvable option: %s", 
		 ((opt->name == 0) ? "(null)" : opt->name));
	}
      opt--;
      goto tail_recurse;
    }
  else
    {
      return opt;
    }
}

/* Return `no_argument', `optional_argument' or `required_argument'
   depending on whether the option `opt' requires or accepts arguments.
   This is infered from opt->arg. */
static int
parg_get_has_arg (struct parg *parser, struct parg_option *opt)
{
  /* Resolve possible aliases */
  /*  opt = parg_resolve_alias (parser, opt); */

  char *arg = (parg_resolve_alias (parser, opt))->arg;
  return ((arg == 0) ? no_argument 
	  : (((arg[0] == '[') && (arg[(strlen (arg))-1] == ']')) 
	   ? optional_argument : required_argument));
}

/* Returm the argument string associated with option 'opt'. This is 
   infered from opt->arg. If the option takes no arguments, it returns
   the empty string. If the option takes a mandatory argument, it returns
   opt->arg. If the option takes an optional argument, it returns 
   opt->arg without the brackets. */

static char *
parg_get_argument (struct parg *parser, struct parg_option *opt)
{
  int has_arg = parg_get_has_arg (parser, opt);
  static char *empty_string = "";
  static char *optional_arg_string = NULL;
  static int optional_arg_string_length = 0;
  
  /* Resolve possible aliases */
  opt = parg_resolve_alias (parser, opt);

  /* Grok the argument */
  if (has_arg == no_argument)
    return empty_string;
  else if (has_arg == required_argument)
    return opt->arg;
  else
    {
      /* Get the length of opt->arg */
      int arg_length = strlen (opt->arg);

      /* If optional_arg_string has not been initialized, or if
	 it does not have enough length, reallocate */
      if (optional_arg_string_length < arg_length)
	{
	  optional_arg_string = REALLOC(optional_arg_string, char, 
					((2*arg_length)+1));
	  optional_arg_string_length = 2*arg_length;
	}

      /* Copy the opt->arg part inside [] to optional_arg_string */
      memcpy (optional_arg_string, (opt->arg)+1, arg_length - 2);
      optional_arg_string[arg_length-2] = '\0';
      return optional_arg_string;
    }
}

/* When getopt_long encounters a long option it returns an index to
   that long option. This function takes the index and finds the 
   struct parg_option entry that corresponds to the long option that has
   been encountered */
struct parg_option *
parg_get_option_from_index (struct parg *parser, int index)
{
  /* Get the name of the option from the index */
  const char *name = parser->longopts[index].name;

  /* Look for the name in the parg parser */
  struct parg_option *opt;
  for (opt = parser->options; parg_terminating_option_p (opt) == 0; opt++)
    {
      if (opt->key == 0) goto skip_this_option;
      if (opt->name != 0) if (strcmp(name,opt->name) == 0) return opt;
    skip_this_option:
      while (0); /* Kludge for losing compilers, such as in Digital Unix */
    }

  /* We shouldn't get to this point */
  error (1, 0, "Internal Error: parg_get_option_from_index barfed\n"
	 "The offending index is: %d", index);
}

/* When getopt_long encounters a short option, it returns a key. 
   This routine takes the key and finds the struct parg_option entry
   that corresponds the the short option encountered */
struct parg_option *
parg_get_option_from_key (struct parg *parser, int key)
{
  /* Look for the key in the parg parser */
  struct parg_option *opt;
  for (opt = parser->options; parg_terminating_option_p (opt) == 0; opt++)
    {
      if (opt->key == 0) goto skip_this_option;
      if (key == opt->key) return opt;
    skip_this_option:
      while (0);
    }
  return 0; 
}

/* Call the hook that corresponds to the option opt. Recurse if necessary */
static void 
parg_run_hook (struct parg *parser, struct parg_option *opt, char *arg)
{
  opt = parg_resolve_alias (parser, opt);
  (*(opt->hook)) (parser, arg);
  return;
}

/* 
 * --------------------------------------------------------------------
 * Parses the command-line switches and dispatches the appropriate
 * callbacks. This routine must be called only after a call to
 * parg_set_invocation.
 * --------------------------------------------------------------------
 */

struct parg *
parg_parse (int main_argc, char **main_argv, struct parg_option *options)
{
  /* Allocate the parser */
  struct parg *parser = MALLOC(struct parg, 1);
  struct parg_option *opt;
  int i,c;
  int option_index;
  parg_option_hook *hook;

  /* Useful enumerations on the options vector */
  int N_option_entries = 0;
  int N_long = 0;
  int N_short_no_arg = 0;
  int N_short_optional_arg = 0;
  int N_short_required_arg = 0;

  /* Copy the arguments */
  parser->argc = main_argc;
  parser->argv = main_argv;
  parser->options = options; 

  /* Unless we passed a NULL options, count them */
  if (options != 0)
    {
      opt = options;
      while (parg_terminating_option_p (opt++) == 0)
	{
	  int has_arg = parg_get_has_arg (parser, opt);
	  N_option_entries++;
	  N_long += parg_long_option_p (opt);
	  if (parg_short_option_p (opt) == 1)
	    {
	      N_short_no_arg += ((has_arg == no_argument) ? 1 : 0);
	      N_short_optional_arg += ((has_arg == optional_argument) ? 1 : 0);
	      N_short_required_arg += ((has_arg == required_argument) ? 1 : 0);
	    }
	}
    }

  /* Take into account the options --help, --usage and --version */
  N_long += 3;
  N_short_no_arg += 2;

  /* Allocate options vectors for getopt_long */
  parser->longopts = CALLOC(struct option, (N_long + 1));
  parser->shortopts = CALLOC(char, (N_short_no_arg + 2*N_short_optional_arg +
				    3*N_short_required_arg));

  /* Entry for --help */
  parser->longopts[0].name = xstrdup ("help");
  parser->longopts[0].has_arg = no_argument;
  parser->longopts[0].flag = 0;
  parser->longopts[0].val = 'h';
  
  /* Entry for --version */
  parser->longopts[1].name = xstrdup ("version");
  parser->longopts[1].has_arg = no_argument;
  parser->longopts[1].flag = 0;
  parser->longopts[1].val = 'V';

  /* Entry for --usage */
  parser->longopts[2].name = xstrdup ("usage");
  parser->longopts[2].has_arg = no_argument;
  parser->longopts[2].flag = 0;
  parser->longopts[2].val = '@';

  /* Corresponding entries for shortopts */
  strcat (parser->shortopts, "hV");

  /* Now loop over the options array and add entries to longopts and
   * shortopts for the user-requested options */
  if (options != 0)
    {
      for (opt = options, i = 3; parg_terminating_option_p (opt) == 0; opt++)
	{
          /* Determine whether this option has an argument */
	  int has_arg = parg_get_has_arg (parser, opt);

          /* If this is a documentation option, then skip it */
	  if (parg_documentation_option_p (opt)) goto skip_this_option;

          /* Set longopts for this option */
	  if (parg_long_option_p (opt))
	    {
	      parser->longopts[i].name = xstrdup (opt->name);
	      parser->longopts[i].has_arg = has_arg;
	      parser->longopts[i].flag = 0;
	      parser->longopts[i].val = (opt->key == ':') ? 0 : opt->key;
	      i++; /* Increment the counter for long options */
	    }
          
          /* Set shortopts for this option */
	  if (parg_short_option_p (opt)) 
	    {
	      char tmp[4] = { '\0', '\0', '\0', '\0' };
	      tmp[0] = (char) opt->key;
	      if (has_arg == required_argument)
		{ tmp[1] = ':'; }
	      else if (has_arg == optional_argument)
		{ tmp[1] = ':'; tmp[2] = ':'; }
	      strcat (parser->shortopts,tmp);
	    }

	skip_this_option:
	  while (0);
	}
    }

  /* Now call getopt_long and do the loop */
  while ((c = getopt_long (main_argc, main_argv, parser->shortopts, 
			   parser->longopts, &option_index)) != -1)
    {
      /* If it is a long option */
      if (c == 0)
	{
	  parg_run_hook (parser, 
			 parg_get_option_from_index (parser, option_index),
			 optarg);
	}
      /* Hook for --help */
      else if (((char) c) == 'h')
	{
	  parg_hook_help (parser, optarg);
	}
      /* Hook for --version */
      else if (((char) c) == 'V')
	{
	  parg_hook_version (parser, optarg);
	}
      /* Hook for --usage */
      else if (((char) c) == '@')
	{
	  parg_hook_usage (parser, optarg);
	}
      /* Unrecognized or problematic option - call usage */
      else if (((char) c) == '?')
	{
	  parg_hook_usage (parser,optarg);
	}
      /* Short option */
      else
	{
	  parg_run_hook (parser, parg_get_option_from_key (parser, c), optarg);
	}
    }

  /* Return the parser in case user wants to do things with it */
  return parser;
}

/*
 * This routine returns the number of command-line arguments that remain
 * to be parsed that are not options. You must give it as argument the
 * parser argument returned from parg_parse
 */

int 
parg_argument_count (struct parg *parser)
{
  return ((parser->argc) - optind);
}

/*
 * This routine returns the nth argument that remains to be parsed that
 * is not an option. You must give it as argument the parser argument
 * returned from parg_parse. The argument count starts from 1,2,3...,n
 * where n is what gets returned from parg_argument_count
 */

char *
parg_argument (struct parg *parser, int n)
{
  if ((n < 1) || (n > ((parser->argc) - optind)))
    {
      parg_error (parser, "Argument %d requested but not available\n", n);
    }
  return (parser->argv[(optind-1)+n]);
}

/*
** =========================================================================
** =========================================================================
*/

/*
** The following routines are responsible for handling the
** --help, --version and --usage options
*/

/* This routine handles the --version option */
void
parg_hook_version (struct parg *parser, char *arg)
{
  /* Display the program's name */
  fprintf (stderr, "%s", program_invocation_short_name);
  
  /* If the program has a unique version number, display it. 
     Otherwise, display the package version number */
  fprintf (stderr, " %s", ((parg_program_version != 0) 
			   ? parg_program_version : parg_package_version));

  /* If necessary, display the package version and name separately */
  if ((parg_program_version != 0) || 
      (strcmp (program_invocation_short_name, parg_package_name) != 0))
    {
      fprintf (stderr, " - %s %s", parg_package_name, parg_package_version);
    }

  /* Throw a newline */
  fprintf (stderr, "\n");
  
  /* Call the hook */
  if (parg_version_hook != 0)
    {
      (*parg_version_hook) ();
    }
  exit (0);
}

/*
** =========================================================================
** =========================================================================
*/

/*
 * Print the usage string. You need to provide a string for the options.
 * This routine will put together invokation and extract argument syntax
 * from the parg_program_arguments global variable. For the --usage 
 * option, you need to construct a detailed string for the options.
 * For the --help option, it is okey to just pass NULL. Then we simply
 * use [OPTION...] as dictated by the GNITS standards.
 */

static void
parg_describe_usage (struct fmtstring *fs, char *options)
{
  /* Set the wrap margin for the Usage strings */
  fmts_set_wrap (fs, 12);

  /* Add the canonical stuff for usage */
  fmts_catenate2 (fs, "Usage: ", program_invocation_short_name);
  fmts_catenate (fs, ((options == NULL) ? "[OPTION...]" : options));

  /* If we have specified usage for program arguments add it in too */
  if (parg_program_arguments != 0)
    {
      char *p = parg_program_arguments;
      char *p2;

      /* Entry point for handling alternative usage recursively */
    another_usage:

      /* See if there is a \n ahead instead of a p */
      p2 = strchr (p,'\n');
      /* If there is one, then dump the substring between p and p2
	 and setup another line starting with or: to display the 
	 alternative usage */
      if (p2 != NULL)
	{
	  fmts_catenate_substring (fs, p, p2-1);
	  fmts_linebreak (fs);
	  fmts_catenate2 (fs, "   or: ", program_invocation_short_name);
	  fmts_catenate (fs, ((options == NULL) ? "[OPTION...]" : options));
	  p = p2 + 1;
	  goto another_usage;
	}
      else
	{
	  fmts_catenate (fs, p);
	}
    }
  fmts_set_wrap (fs, 0);
}

/*
 * Create a more detailed [OPTIONS] string to use in invoking 
 * parg_describe_usage. This is mainly used in the --usage option 
 */

static char *
parg_get_usage_substring (struct parg *parser)
{
  struct parg_option *options = parser->options;
  dynamic_string *usage_string = MALLOC(dynamic_string, 1);
  ds_init (usage_string, 100);

  /* If there are options to process, do so */
  if (options != 0)
    {
      struct parg_option *opt;
      int have_options_with_no_arguments = 0;

      /* Lump together the short options that take no arguments */
      for (opt = options; parg_terminating_option_p (opt) == 0; opt++)
	{
	  if ((parg_short_option_p (opt)) &&
	      (parg_get_has_arg (parser, opt) == no_argument))
	    {
	      /* A temporary one character string containing the option */
	      char tmp[2] = {'\0', '\0'};
	      tmp[0] = opt->key;

	      /* If this is the first such option, then open the brace */
	      if (have_options_with_no_arguments == 0) 
		{
		  ds_strcat (usage_string, "[-"); 
		  have_options_with_no_arguments = 1;
		}

	      /* Output the letter for this option */
	      ds_strcat (usage_string, tmp);
	    }
	}

      /* If there were short options taking no arguments, then we need to
	 close the bracket */
      if (have_options_with_no_arguments) ds_strcat (usage_string, "] ");

      /* Now process the short options with arguments */
      for (opt = options; parg_terminating_option_p (opt) == 0; opt++)
	{
	  if (parg_short_option_p (opt))
	    {
	      int has_arg = parg_get_has_arg (parser, opt);
	      char *arg = parg_get_argument (parser, opt);
	      char tmp[2] = {'\0', '\0'};
	      tmp[0] = opt->key;

	      /* If the option takes an argument, then add it in */
	      if (has_arg != no_argument)
		{
		  ds_strcat (usage_string, "[-");
		  ds_strcat (usage_string, tmp);
		  /* How we present the argument depends on whether it is
		     optional or required */
		  if (has_arg == optional_argument)
		    {
		      ds_strcat (usage_string, "[");
		      ds_strcat (usage_string, arg);
		      ds_strcat (usage_string, "]");
		    }
		  else if (has_arg == required_argument)
		    {
		      ds_strcat (usage_string, "\n");
		      ds_strcat (usage_string, arg);
		    }
		  /* Close the brace */
		  ds_strcat (usage_string, "] ");
		}
	    }
	}

      /* Now, process the long options, regardless of whether or not
	 they take arguments  all in one pass */
      for (opt = options; parg_terminating_option_p (opt) == 0; opt++)
	{
	  if (parg_long_option_p (opt))
	    {
	      int has_arg = parg_get_has_arg (parser, opt);
	      ds_strcat (usage_string, "[--");
	      ds_strcat (usage_string, opt->name);
	      if (has_arg == optional_argument)
		{
		  ds_strcat (usage_string, "=[");
		  ds_strcat (usage_string, parg_get_argument (parser, opt));
		  ds_strcat (usage_string, "]");
		}
	      else if (has_arg == required_argument)
		{
		  ds_strcat (usage_string,"=");
		  ds_strcat (usage_string, parg_get_argument (parser, opt));
		}
	      ds_strcat (usage_string, "] ");
	    }
	}
    }

  /* Finally, add entries for --help, --usage, --version */
  ds_strcat (usage_string, "[--help] [--usage] [--version]");
  return ds_get_string (usage_string);  
}

static
void parg_print_usage (struct parg *parser)
{
  struct fmtstring *fs = new_fmtstring (0, 78, 0);
  parg_describe_usage (fs, parg_get_usage_substring (parser));
  fprintf (stderr, fmts_string (fs));
  fprintf (stderr, "\n");
}

void 
parg_hook_usage (struct parg *parser, char *arg)
{
  parg_print_usage (parser);
  exit (0);
}

/*
** =======================================================================
** =======================================================================
*/


/* Add ONE entry to the long listing of options and their documentation
   string to the formatted string. This assumes that you have accumulated
   information on both the option and its aliases to the short_options and
   long_option string. You must also provide doc_string */
static void
parg_describe_an_option (struct fmtstring *fs, char *short_options,
			 char *long_options, char *doc_string)
{
#if 0
  fprintf (stderr, "short_options = %s\n", short_options);
  fprintf (stderr, "long_options = %s\n", long_options);
  fprintf (stderr, "doc_string = %s\n", doc_string);
#endif

  /* The following constants configure how we want to do the indentation */
  const int short_option_indent = 2;
  const int long_option_indent = 6;
  const int doc_string_indent = 29;
  const int option_right_margin = 28;
  const int option_wrap_margin = 6;

  /* Modify margins for outputing the option information */
  fmts_set_left (fs, short_option_indent);
  fmts_set_right (fs, option_right_margin);
  fmts_set_wrap (fs, option_wrap_margin);

  /* If we have short options, bring them out, otherwise go to 
     long_option_indent  */
  if (strlen (short_options) > 0)
    {
      if (strlen (long_options) > 0)
	{
	  fmts_catenate3 (fs, short_options, ", ", long_options); 
	}
      else
	{
	  fmts_catenate (fs, short_options);
	}
    }
  else
    {
      fmts_move_to_column (fs, long_option_indent);
      fmts_catenate (fs, long_options);
    }

  /* Print the documentation string */
  fmts_move_to_column (fs, doc_string_indent);
  fmts_set_left (fs, doc_string_indent);
  fmts_set_right (fs, 79);
  fmts_set_wrap (fs, doc_string_indent);
  fmts_catenate (fs, doc_string);

  /* Get a linebreak */
  fmts_set_left (fs, 0); 
  fmts_set_right (fs, 79); 
  fmts_set_wrap (fs, 0);
  fmts_linebreak (fs);
}

/* Add a long listing of the options and their documentation string
   to the formatted string fs */
static void
parg_describe_options (struct fmtstring *fs, struct parg *parser)
{
  dynamic_string *short_option_string = MALLOC(dynamic_string, 1);
  dynamic_string *long_option_string = MALLOC(dynamic_string, 1);
  char *doc_string = 0;
  struct parg_option *opt;

  for (opt = parser->options; parg_terminating_option_p (opt) == 0; opt++)
    {
      /* If the current option is a documentation option, then handle it
	 as such */
      if (parg_documentation_option_p (opt))
	{
	  fmts_linebreak (fs);
	  fmts_move_to_column (fs, 1);
	  fmts_catenate (fs, opt->name);
	  fmts_linebreak (fs);
	}
      /* If this is a real option, but not an alias then handle it */
      else if (parg_option_is_alias_p (opt) == 0)
	{
	  /* If the documentation string is null (and it is not a 
	     documentation option) then skip processing this option.
	     Otherwise keep a pointer to the documentation string. */
	  if (opt->doc == 0) 
	    goto skip_this_option;
	  else 
	    doc_string = opt->doc;


	  /* Setup the option strings */
	  ds_init (short_option_string, 80);
	  ds_init (long_option_string, 80);

	  /* Recursion jump point for handling alias options */
	handle_the_option:

	  /* Handle, if it is a short option */
	  if (parg_short_option_p (opt))
	    {
	      char tmp[2] = {'\0', '\0'};
	      tmp[0] = (char) opt->key;
	      ds_strcat (short_option_string, "-");
	      ds_strcat (short_option_string, tmp);
	    }
	  
	  /* Handle, if it is a long option */
	  if (parg_long_option_p (opt))
	    {
	      /* Two slashes and the name of the option */
	      int has_arg = parg_get_has_arg (parser, opt);
	      ds_strcat (long_option_string, "--");
	      ds_strcat (long_option_string, opt->name);
	      /* Indicate the arguments, unless the next option is an
		 alias to this option */
	      if ((parg_terminating_option_p (opt+1) == 0) &&
		  (parg_option_is_alias_p (opt+1) == 0))
		{
		  if (has_arg == optional_argument)
		    {
		      ds_strcat (long_option_string, "[=");
		      ds_strcat (long_option_string, 
				 parg_get_argument (parser, opt));
		      ds_strcat (long_option_string,"]");
		    }
		  else if (has_arg == required_argument)
		    {
		      ds_strcat (long_option_string, "=");
		      ds_strcat (long_option_string,
				 parg_get_argument (parser, opt));
		    }
		}
	    }

	  /* If the subsequent option is not the terminating option,
	     then we should check whether it is an alias to this option */
	  if (parg_terminating_option_p (opt+1) == 0)
	    {
	      /* If the option is an alias we must process it now! */
	      if (parg_option_is_alias_p (opt+1))
		{
		  /* Make the alias the current option */
		  opt++;

		  /* Depending on whether the option is long or short, add
		     a comma to the corresponding dynamic *string */
		  if (parg_short_option_p (opt)) 
		    ds_strcat (short_option_string, ", "); 
		  if (parg_long_option_p (opt))
		    ds_strcat (long_option_string, ", ");
		  
		  /* Now recurse to handle the alias option */
		  goto handle_the_option;
		}
	    }

	  /* Okey, now we have all the strings ready, punch them in */
	  parg_describe_an_option (fs, ds_get_string (short_option_string),
				   ds_get_string (long_option_string), 
				   doc_string);
	  ds_destroy (short_option_string);
	  ds_destroy (long_option_string);

	  /* Jump point for skipping a hidden option */
	skip_this_option:
	  while (0);
	}
      
    }
}


void
parg_hook_help (struct parg *parser, char *arg)
{
  /* Allocate a formatted string */
  struct fmtstring *fs = new_fmtstring (0, 78, 0);

  /* Add the usage information */
  parg_describe_usage (fs, NULL);

  /* Two line breaks between usage and documentation */
  fmts_linebreak (fs);
  fmts_linebreak (fs);

  /* Display the first part of the documentation */
  if (parg_program_documentation != 0)
    {
      /* Look for a separating \v character */
      char *p = strchr (parg_program_documentation,'\v');
      /* If there is no \v character put in the whole string */
      if (p == NULL)
	{
	  fmts_catenate (fs, parg_program_documentation);
	  fmts_linebreak (fs); fmts_linebreak (fs);
	  parg_program_documentation = NULL;;
	}
      /* If there is a \v character, put in only until \v and set
	 parg_program_documentation to point to the second part of the 
	 string */
      else
	{
	  fmts_catenate_substring (fs, parg_program_documentation, p-1);
	  fmts_linebreak (fs); fmts_linebreak (fs);
	  parg_program_documentation = p+1;
	}
    }

  /* Describe the standard --help, --version, --usage options */
  parg_describe_an_option (fs, "-h", "--help", "Give this help list");
  parg_describe_an_option (fs, "", "--usage", "Give a short usage message");
  parg_describe_an_option (fs, "-V", "--version", "Print program version");

  /* Display help for the options. This is lots of work so we do it
     in a separate subroutine */
  if (parser->options != 0) parg_describe_options (fs, parser);
  
  /* Add a line break */
  fmts_linebreak (fs);

  /* If we have any long options with arguments, then print this 
     very much debated boilerplate note */
  if (parser->options != 0)
  {
    int print_statement = 0;
    struct parg_option *opt;
    for (opt = parser->options; parg_terminating_option_p (opt) == 0; opt++)
      {
	if (parg_documentation_option_p (opt) == 0)
	  {
	    int has_arg = parg_get_has_arg (parser, opt);
	    if ((has_arg == required_argument) || 
		(has_arg == optional_argument))
	      print_statement = 1;
	  }
      }
    if (print_statement)
      {
	fmts_catenate (fs, "Mandatory or optional arguments to long options "
		       "are also mandatory or optional for any corresponding "
		       "short options.");
	fmts_linebreak (fs);
	fmts_linebreak (fs);
      }
  }

  /* Display the last part of the documentation */
  if (parg_program_documentation != 0)
    {
      fmts_catenate (fs, parg_program_documentation);
      fmts_linebreak (fs); fmts_linebreak (fs);
    }

  
  /* Bug reports */
  if (parg_program_bug_address != 0)
    {
      fmts_catenate2 (fs, "Please report bugs to ", parg_program_bug_address);
    }

  /* Done constructing the help string. Now print it */
  fprintf (stderr, fmts_string (fs));
  fprintf (stderr, "\n");
  exit (0);
}

/*
** ========================================================================
** ========================================================================
*/

/* This is a little messy, but this is the GNITS way of writing
   variadic functions. Oh well... */

#if HAVE_VPRINTF || HAVE_DOPRNT || _LIBC
# if __STDC__
#  include <stdarg.h>
#  define VA_START(args, lastarg) va_start(args, lastarg)
# else
#  include <varargs.h>
#  define VA_START(args, lastarg) va_start(args)
# endif
#else
# define va_alist a1, a2, a3, a4, a5, a6, a7, a8
# define va_dcl char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
#endif

void
#if defined VA_START && __STDC__
parg_error (struct parg *parser, const char *fmt, ...)
#else
parg_error (parser, fmt, va_alist)
     struct parg *parser;
     const char *fmt;
     va_dcl
#endif
{
#ifdef VA_START
  va_list args;
#endif

  /* Print the program invocation name */
  fprintf (stderr, "%s:", program_invocation_name);

  /* Call vprintf to print the desired error message */
#ifdef VA_START
  VA_START (args, fmt);
# if HAVE_VPRINTF || _LIBC
  vfprintf (stderr, fmt, args);
# else
  _doprnt (message, args, stderr);
# endif
  va_end (args);
#else
  fprintf (stderr, fmt, a1, a2, a3, a4, a5, a6, a7, a8);
#endif

  /* Print the --usage string */
  fprintf (stderr, "\n");
  parg_print_usage (parser);

  /* Print the try for help message */
  fprintf (stderr, "Try %s --help for more information\n",
	   program_invocation_name);
  exit (1);
}

/*
** =======================================================================
** =======================================================================
*/

/* The following functions are useful for parsing arguments that are
   actually integers or numbers */

#include "xstrtol.h"
#include "xstrtod.h"

/* If the following global variables is set to zero, then 
   the parsing functions print warning and not fatal errors.
   This is a useful mode when you are parsing interactive input
   from the user. For parsing command-line arguments, the default
   is to exit the program */
int parg_conversions_kill_program_p = 1;

/* Take a string containing an integer and return the integer */
long int parg_string_to_integer (char *str)
{
  long int val;
  char *p;
  strtol_error s_err = xstrtol (str, &p, 0, &val, "bckmw");
  if (s_err == LONGINT_OK)
    {
      return val;
    }
  else
    {
      if (parg_conversions_kill_program_p)
	{ 
	  STRTOL_FATAL_ERROR (str, "arg", s_err); 
	}
      else
	{ 
	  STRTOL_FAIL_WARN (str, "arg", s_err); 
	}
    }
}

/* Take a string containing a real number and return the number */
double parg_string_to_double (char *str)
{
  double val;
  
  if (xstrtod (str, 0, &val) == 0)
    {
      return val;
    }
  else
    {
      error (parg_conversions_kill_program_p, 0, 
	     "invalid real number %s", str);
    }
}
