/*
** Copyright (C) 1998 Eleftherios Gkioulekas <lf@amath.washington.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifndef __defined_parg_h
#define __defined_parg_h

/* Necessary because we declare variadic functions */
#ifdef __cplusplus
extern "C" {
#endif

struct parg;
struct parg_option;
typedef int parg_option_hook (struct parg *parser, char *arg);

/*  
** The following structure holds together the list of options that
** are understood by the program
*/

struct parg_option
{
  /* The long name of the option. If this is a documentation option, then
   * this is the content of the documentation option */

  char *name; 

  /* The short name of the option, if the option has one.
   * ':' if the option does not have a short name
   * 0 if it is a documentation option
   */

  int key; 

  /* If NULL, then the option takes no arguments. 
   * If non-NULL, then the option may take an argument whose name is
   * the value of arg. If the name is surrounded in [] then it is optional
   * otherwise it is required.
   */

  char *arg; 

  /* Documentation string for the option explaining in one line what the 
   * option does. If the string is 0 then the option is an alias of the
   * previous option and the hook must also be set equal to 0! */

  char *doc;

  /* Hook to call when an option is encountered. If the option takes an
   * argument then it is passed on to the hook. If the hook is 0 but key
   * or name are not 0, then this is an alias of a previous option */

  parg_option_hook *hook;

};

/*
** The following structure holds together the state of the parser
*/

struct parg
{
  /* The argument vector being parsed. May be modified */
  int argc;
  char **argv;

  /* The options vector being parsed */
  struct parg_option *options;

  /* The options vectors passed to getopt */
  struct option *longopts;
  char *shortopts;
 
};

/*
** Public methods associated with the parser
*/

/* Global variables */
extern char *parg_program_bug_address;
extern char *parg_package_name;
extern char *parg_package_version;
extern char *parg_program_name;
extern char *parg_program_version;
extern char *parg_program_documentation;
extern char *parg_program_arguments;
extern void (*parg_version_hook) (void);

/* The GNU invocation global variables */
extern char *program_invocation_name;
extern char *program_invocation_short_name;

/* The main entry points */
void parg_set_invocation (char *argv0, char *package, char *version);
struct parg * parg_parse (int main_argc, char **main_argv, 
			  struct parg_option *options);
int parg_argument_count (struct parg *parser);
char *parg_argument (struct parg *parser, int i);

/* The error function.  */
#if defined(__STDC__) && __STDC__
void parg_error (struct parg *parser, const char *fmt, ...);
#else
void parg_error ();
#endif

/* Parse arguments that are integers or real numbers */
extern int parg_conversions_kill_program_p;
long int parg_string_to_integer (char *str);
double parg_string_to_double (char *str);

/*
** Macros for generating standard hooks 
*/

/* The flag hook creates hooks for handling options that take no arguments
   that simply want to set a flag */

#define PARG_FLAG_HOOK(name) 				\
int name ## _p = 0; 					\
int name ## _hook (struct parg *parser, char *arg) 	\
{ 							\
  name ## _p = 1; 					\
} 

/* This flag creates a hook for handling options with required arguments.
   If the option appears, then name_p is set equal to 1, and name_argument
   contains the the required argument */

#define PARG_GET_ARG_HOOK(name)				\
int name ## _p = 0;					\
char * name ## _argument = 0;				\
int name ## _hook (struct parg *parser, char *arg)	\
{							\
  name ## _p = 1;					\
  name ## _argument = arg;				\
}

/* This flag creates a hook for handling options with optional arguments.
   It allows the developer to provide a default value to be used if the
   argument is not given. Similarly, name_p is set equal to 1 if the option
   appears but name_argument is guaranteed to always have the Right Value */

#define PARG_GET_OPTIONAL_ARG_HOOK(name, default)	\
int name ## _p = 0;					\
char * name ## _argument = default;			\
int name ## _hook (struct parg *parser, char *arg)	\
{							\
  name ## _p = 1;					\
  if (arg != 0) name ## _argument = arg;		\
}

#endif



/* End extern "C" */
#ifdef __cplusplus
}
#endif
